# -*- coding: utf-8 -*-
"""
This class allow to transform SQLAlchemy metadata to the intermediary syntax.
"""

from eralchemy.models import Relation, Column, Table
import sys
from sqlalchemy.exc import CompileError

if sys.version_info[0] == 3:
    unicode = str


def relation_to_intermediary(fk):
    """Transform an SQLAlchemy ForeignKey object to it's intermediary representation. """
    return Relation(
        right_col=format_name(fk.parent.table.fullname),
        left_col=format_name(fk._column_tokens[1]),
        right_cardinality='?',
        left_cardinality='*',
    )


def format_type(typ):
    """ Transforms the type into a nice string representation. """
    try:
        return unicode(typ)
    except CompileError:
        return 'Null'


def format_name(name):
    """ Transforms the name into a nice string representation. """
    return unicode(name)


def column_to_intermediary(col, type_formatter=format_type):
    """Transform an SQLAlchemy Column object to it's intermediary representation. """
    return Column(
        name=col.name,
        type=type_formatter(col.type),
        is_key=col.primary_key,
    )


def table_to_intermediary(table):
    """Transform an SQLAlchemy Table object to it's intermediary representation. """
    return Table(
        name=table.fullname,
        columns=[column_to_intermediary(col) for col in table.c._data.values()]
    )


def metadata_to_intermediary(metadata):
    """ Transforms SQLAlchemy metadata to the intermediary representation. """
    tables = [table_to_intermediary(table) for table in metadata.tables.values()]
    relationships = [relation_to_intermediary(fk) for table in metadata.tables.values() for fk in table.foreign_keys]
    return tables, relationships


def declarative_to_intermediary(base):
    """ Transform an SQLAlchemy Declarative Base to the intermediary representation. """
    return metadata_to_intermediary(base.metadata)


def database_to_intermediary(database_uri, schema=None):
    """ Introspect from the database (given the database_uri) to create the intermediary representation. """
    from sqlalchemy.ext.automap import automap_base
    from sqlalchemy import create_engine

    Base = automap_base()
    engine = create_engine(database_uri)
    if schema is not None:
        Base.metadata.schema = schema

    # reflect the tables
    Base.prepare(engine, reflect=True)
    return declarative_to_intermediary(Base)



# -*- coding: utf-8 -*-
from eralchemy.version import version as __version__
from eralchemy.cst import GRAPH_BEGINNING
from eralchemy.sqla import metadata_to_intermediary, declarative_to_intermediary, database_to_intermediary
from pygraphviz.agraph import AGraph
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import ArgumentError
from eralchemy.helpers import check_args
from eralchemy.parser import markdown_file_to_intermediary, line_iterator_to_intermediary, ParsingException
import argparse
import sys

try:
    basestring
except NameError:
    basestring = str


def cli():
    """Entry point for the application script"""
    parser = argparse.ArgumentParser(prog='ERAlchemy')
    parser.add_argument('-i', nargs='?', help='Database URI to process.')
    parser.add_argument('-o', nargs='?', help='Name of the file to write.')
    parser.add_argument('-s', nargs='?', help='Name of the schema.')
    parser.add_argument('-x', nargs='*', help='Name of the table(s) to exclude.')
    parser.add_argument('-v', help='Prints version number.', action='store_true')

    args = parser.parse_args()
    check_args(args)
    if args.v:
        print('ERAlchemy version {}.'.format(__version__))
        exit(0)
    render_er(args.i, args.o, exclude=args.x, schema=args.s)


def intermediary_to_markdown(tables, relationships, output):
    """ Saves the intermediary representation to markdown. """
    er_markup = _intermediary_to_markdown(tables, relationships)
    with open(output, "w") as file_out:
        file_out.write(er_markup)


def intermediary_to_dot(tables, relationships, output):
    """ Save the intermediary representation to dot format. """
    dot_file = _intermediary_to_dot(tables, relationships)
    with open(output, "w") as file_out:
        file_out.write(dot_file)


def intermediary_to_schema(tables, relationships, output):
    """ Transforms and save the intermediary representation to the file chosen. """
    dot_file = _intermediary_to_dot(tables, relationships)
    graph = AGraph()
    graph = graph.from_string(dot_file)
    extension = output.split('.')[-1]
    graph.draw(path=output, prog='dot', format=extension)


def _intermediary_to_markdown(tables, relationships):
    """ Returns the er markup source in a string. """
    t = '\n'.join(t.to_markdown() for t in tables)
    r = '\n'.join(r.to_markdown() for r in relationships)
    return '{}\n{}'.format(t, r)


def _intermediary_to_dot(tables, relationships):
    """ Returns the dot source representing the database in a string. """
    t = '\n'.join(t.to_dot() for t in tables)
    r = '\n'.join(r.to_dot() for r in relationships)
    return '{}\n{}\n{}\n}}'.format(GRAPH_BEGINNING, t, r)

# Routes from the class name to the function transforming this class in
# the intermediary representation.
switch_input_class_to_method = {
    'MetaData': metadata_to_intermediary,
    'DeclarativeMeta': declarative_to_intermediary,
    # For compatibility with Flask-SQLAlchemy
    '_BoundDeclarativeMeta': declarative_to_intermediary
}

# Routes from the mode to the method to transform the intermediary
#  representation to the desired output.
switch_output_mode_auto = {
    'er': intermediary_to_markdown,
    'graph': intermediary_to_schema,
    'dot': intermediary_to_dot
}

# Routes from the file extension to the method to transform
# the intermediary representation to the desired output.
switch_output_mode = {
    'er': intermediary_to_markdown,
    'dot': intermediary_to_dot,
}


def all_to_intermediary(filename_or_input, schema=None):
    """ Dispatch the filename_or_input to the different function to produce the intermediary syntax.
    All the supported classes names are in `swich_input_class_to_method`.
    The input can also be a list of strings in markdown format or a filename finishing by '.er' containing markdown
    format.
    """
    # Try to convert from the name of the class
    input_class_name = filename_or_input.__class__.__name__
    try:
        this_to_intermediary = switch_input_class_to_method[input_class_name]
        tables, relationships = this_to_intermediary(filename_or_input)
        return tables, relationships
    except KeyError:
        pass

    # try to read markdown file.
    if isinstance(filename_or_input, basestring):
        if filename_or_input.split('.')[-1] == 'er':
            return markdown_file_to_intermediary(filename_or_input)

    # try to read a markdown in a string
    if not isinstance(filename_or_input, basestring):
        if all(isinstance(e, basestring) for e in filename_or_input):
            return line_iterator_to_intermediary(filename_or_input)

    # try to read DB URI.
    try:
        make_url(filename_or_input)
        return database_to_intermediary(filename_or_input, schema=schema)
    except ArgumentError:
        pass

    msg = 'Cannot process filename_or_input {}'.format(input_class_name)
    raise ValueError(msg)


def get_output_mode(output, mode):
    """
    From the output name and the mode returns a the function that will transform the intermediary
    representation to the output.
    """
    if mode != 'auto':
        try:
            return switch_output_mode_auto[mode]
        except KeyError:
            raise ValueError('Mode "{}" is not supported.')

    extension = output.split('.')[-1]
    try:
        return switch_output_mode[extension]
    except KeyError:
        return intermediary_to_schema


def filter_excludes(tables, relationships, exclude):
    """ This function excludes the tables and relationships with tables which are
    in the exclude (lst of str, tables names) """
    exclude = exclude or []
    tables_filtered = [t for t in tables if t.name not in exclude]
    relationships_filtered = [r for r in relationships
                              if r.right_col not in exclude and r.left_col not in exclude]
    return tables_filtered, relationships_filtered


def render_er(input, output, mode='auto', exclude=None, schema=None):
    """
    Transforms the metadata into a representation.
    :param input: Possible inputs are instances of:
        MetaData: SQLAlchemy Metadata
        DeclarativeMeta: SQLAlchemy declarative Base
    :param output: name of the file to output the
    :param mode: str in list:
        'er': writes to a file the markup to generate an ER style diagram.
        'graph': writes the image of the ER diagram.
        'dot': write to file the diagram in dot format.
        'auto': choose from the filename:
            '*.er': writes to a file the markup to generate an ER style diagram.
            '.dot': returns the graph in the dot syntax.
            else: return a graph to the format graph
    :param schema: name of the schema
    """
    try:
        tables, relationships = all_to_intermediary(input, schema=schema)
        if exclude is not None:
            tables, relationships = filter_excludes(tables, relationships, exclude)
        intermediary_to_output = get_output_mode(output, mode)
        intermediary_to_output(tables, relationships, output)
    except ImportError as e:
        module_name = e.message.split()[-1]
        print('Please install {0} using "pip install {0}".'.format(module_name))
    except ParsingException as e:
        sys.stderr.write(e.message)


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
from eralchemy.cst import TABLE, FONT_TAGS, ROW_TAGS
import operator
import re
"""
All the intermediary syntax.
We can several kinds of models can be translated to this syntax.
"""


class Drawable:
    """ Abstract class to represent all the objects which are drawable."""
    RE = None

    def to_markdown(self):
        """Transforms the intermediary object to it's syntax in the er markup. """
        raise NotImplemented()

    def to_dot(self):
        """Transforms the intermediary object to it's syntax in the dot format. """
        raise NotImplemented()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    @staticmethod
    def make_from_match(self):
        """ Used in the parsing of files. Transforms a regex match to a Drawable object. """
        raise NotImplemented()

    def __str__(self):
        return self.to_markdown()


class Column(Drawable):
    """ Represents a Column in the intermediaty syntax """
    RE = re.compile('(?P<primary>\*?)(?P<name>[^\s]+)\s*(\{label:\s*"(?P<label>[^"]+)"\})?')

    @staticmethod
    def make_from_match(match):
        return Column(
            name=match.group('name'),
            type=match.group('label'),
            is_key='*' in match.group('primary'),  # TODO foreign key
        )

    def __init__(self, name, type=None, is_key=False):
        """
        :param name: (str) Name of the column
        :param type:
        :param is_key:
        :return:
        """
        self.name = name
        self.type = type
        self.is_key = is_key

    @property
    def key_symbol(self):
        return '*' if self.is_key else ''

    def to_markdown(self):
        return '    {}{} {{label:"{}"}}'.format(self.key_symbol, self.name, self.type)

    def to_dot(self):
        base = ROW_TAGS.format(' ALIGN="LEFT"', '{key_opening}{col_name}{key_closing}{type}')
        return base.format(
            key_opening='<u>' if self.is_key else '',
            key_closing='</u>' if self.is_key else '',
            col_name=FONT_TAGS.format(self.name),
            type=FONT_TAGS.format(' [{}]').format(self.type) if self.type is not None else ''
        )


class Relation(Drawable):
    """ Represents a Relation in the intermediaty syntax """
    RE = re.compile('(?P<left_name>[^\s]+)\s*(?P<left_cardinality>[*?+1])--(?P<right_cardinality>[*?+1])\s*(?P<right_name>[^\s]+)')
    cardinalities = {
        '*': '0..N',
        '?': '{0,1}',
        '+': '1..N',
        '1': '1',
        '': None
    }

    @staticmethod
    def make_from_match(match):
        return Relation(
            right_col=match.group('right_name'),
            left_col=match.group('left_name'),
            right_cardinality=match.group('right_cardinality'),
            left_cardinality=match.group('left_cardinality'),
        )

    def __init__(self, right_col, left_col, right_cardinality=None, left_cardinality=None):
        if right_cardinality not in self.cardinalities.keys()\
           or left_cardinality not in self.cardinalities.keys():
            raise ValueError('Cardinality should be in {}"'.format(self.cardinalities.keys()))
        self.right_col = right_col
        self.left_col = left_col
        self.right_cardinality = right_cardinality
        self.left_cardinality = left_cardinality

    def to_markdown(self):
        return "{} {}--{} {}".format(
            self.left_col,
            self.left_cardinality,
            self.right_cardinality,
            self.right_col,
        )

    def graphviz_cardinalities(self, card):
        if card == '':
            return ''
        return 'label=<<FONT>{}</FONT>>'.format(self.cardinalities[card])

    def to_dot(self):
        if self.right_cardinality == self.left_cardinality == '':
            return ''
        cards = []
        if self.left_cardinality != '':
            cards.append('tail' +
                         self.graphviz_cardinalities(self.left_cardinality))
        if self.right_cardinality != '':
            cards.append('head' +
                         self.graphviz_cardinalities(self.right_cardinality))
        return '"{}" -- "{}" [{}];'.format(self.left_col, self.right_col, ','.join(cards))

    def __eq__(self, other):
        if Drawable.__eq__(self, other):
            return True
        other_inversed = Relation(
            right_col=other.left_col,
            left_col=other.right_col,
            right_cardinality=other.left_cardinality,
            left_cardinality=other.right_cardinality,
        )
        return other_inversed.__dict__ == self.__dict__


class Table(Drawable):
    """ Represents a Table in the intermediaty syntax """
    RE = re.compile('\[(?P<name>[^]]+)\]')

    def __init__(self, name, columns):
        self.name = name
        self.columns = columns

    @staticmethod
    def make_from_match(match):
        return Table(
            name=match.group('name'),
            columns=[],
        )

    @property
    def header_markdown(self):
        return '[{}]'.format(self.name)

    def to_markdown(self):
        return self.header_markdown + '\n' + '\n'.join(c.to_markdown() for c in self.columns)

    @property
    def header_dot(self):
        return ROW_TAGS.format('', '<B><FONT POINT-SIZE="16">{}</FONT></B>').format(self.name)

    def to_dot(self):
        body = ''.join(c.to_dot() for c in self.columns)
        return TABLE.format(self.name, self.header_dot, body)

    def __str__(self):
        return self.header_markdown

    def __eq__(self, other):
        if not isinstance(other, Table):
            return False
        if other.name != self.name:
            return False

        if sorted(self.columns, key=operator.attrgetter('name')) != sorted(other.columns, key=operator.attrgetter('name')):
            return False
        return True


# -*- coding: utf-8 -*-

"""
All the constants used in the module.
"""
TABLE = '"{}" [label=<<FONT FACE="Helvetica"><TABLE BORDER="0" CELLBORDER="1"' \
        ' CELLPADDING="4" CELLSPACING="0">{}{}</TABLE></FONT>>];'

START_CELL = '<TR><TD ALIGN="LEFT"><FONT>'
FONT_TAGS = '<FONT>{}</FONT>'
# Used for each row in the table.
ROW_TAGS = '<TR><TD{}>{}</TD></TR>'
GRAPH_BEGINNING = (' graph {\n'
                   '    graph [rankdir=LR];\n'
                   '    node [label=\"\\N\",\n'
                   '        shape=plaintext\n'
                   '    ];\n'
                   '    edge [color=gray50,\n'
                   '        minlen=2,\n'
                   '        style=dashed\n'
                   '    ];\n'
                   )